from __future__ import annotations
from collections import defaultdict
import json
from pathlib import Path
import networkx
from networkx import DiGraph
from enum import Enum, auto


class InvalidRelationError(ValueError):
    pass


class Relation(Enum):
    FATHER = auto()
    MOTHER = auto()
    SON = auto()
    DAUGHTER = auto()
    HUSBAND = auto()
    WIFE = auto()

    def __repr__(self) -> str:
        return str(self)


class Person:
    def __init__(self, digraph: DiGraph, id: int, name: str, gender: str) -> None:
        if gender not in ("m", "f"):
            raise ValueError("Gender should be either male(m) or female(f)")
        self.__graph = digraph
        self.__id = id
        self.name = name
        self.__gender = gender[0].lower()
        self.__relatives_dict: dict[Relation, list[Person]] = defaultdict(list)

    @property
    def id(self) -> int:
        return self.__id

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name.title()

    @property
    def gender(self) -> str:
        return self.__gender

    @property
    def parents(self) -> list[Person]:
        relatives = self.relatives_dict()
        return [*relatives[Relation.FATHER], *relatives[Relation.MOTHER]]

    @property
    def father(self) -> Person | None:
        father_list = self.relatives_dict()[Relation.FATHER]
        if father_list:
            return father_list[0]
        return None

    @property
    def mother(self) -> Person | None:
        mother_list = self.relatives_dict()[Relation.MOTHER]
        if mother_list:
            return mother_list[0]
        return None

    @property
    def children(self) -> list[Person]:
        relatives = self.relatives_dict()
        return [*relatives[Relation.SON], *relatives[Relation.DAUGHTER]]

    @property
    def sons(self) -> list[Person]:
        return self.relatives_dict()[Relation.SON]

    @property
    def daughters(self) -> list[Person]:
        return self.relatives_dict()[Relation.DAUGHTER]

    @property
    def husband(self) -> list[Person]:
        return self.relatives_dict()[Relation.HUSBAND]

    @property
    def wife(self) -> list[Person]:
        return self.relatives_dict()[Relation.WIFE]

    def relation_with(self, relative: Person) -> Relation | None:
        try:
            return self.__graph.edges[self, relative][Relation]
        except KeyError:
            return

    def relatives_dict(self) -> dict[Relation, list[Person]]:
        return self.__relatives_dict

    def __add_relation(self, to: Person, relation: Relation) -> None:
        if self is to:
            raise InvalidRelationError("Can't be related to self")
        if self.relation_with(to) is not None:
            raise InvalidRelationError(
                f"Relation is already present ({self.relation_with(to)})"
            )

        self.__graph.add_edges_from([(self, to, {Relation: relation})])
        self.__relatives_dict[relation].append(to)

    def remove_relative(self, relative: Person) -> None:
        def remove_from_one_side(person1: Person, person2: Person):
            relation: Relation = person1.relation_with(person2)
            if relation is not None:
                person1.__relatives_dict[relation].remove(person2)
                person1.__graph.remove_edge(person1, person2)
            else:
                raise InvalidRelationError("Relation not present")

        remove_from_one_side(self, relative)
        remove_from_one_side(relative, self)

    def self_remove(self):
        person = self

        # Removing this person from relatives' reltives_dict
        # This can be done using self.remove_relative(relative), but it removes from both sides, which is not required here, since the person object will be deleted
        for _, relatives in person.relatives_dict().items():
            for relative in relatives:
                relative.__relatives_dict[relative.relation_with(person)].remove(person)

        person.__graph.remove_node(person)

    @staticmethod
    def __validate_is_Person_object(x):
        if not isinstance(x, Person):
            raise ValueError(f"{x} is not a 'Person' object")

    def add_child(self, child: Person) -> None:
        self.__validate_is_Person_object(child)

        child_rel = Relation.SON if child.__gender == "m" else Relation.DAUGHTER
        parent_rel = Relation.FATHER if self.__gender == "m" else Relation.MOTHER

        if self.relation_with(child) is not None:
            raise InvalidRelationError(
                f"Relation is already present ({self.relation_with(child)})"
            )

        if len(child.relatives_dict()[parent_rel]) >= 1:
            raise InvalidRelationError("Can't have multiple father or mother values")

        if parent_rel == Relation.FATHER:
            if (
                child.mother is not None
                and self.relation_with(child.mother) != Relation.WIFE
            ):
                raise InvalidRelationError(
                    f"{self} and {child.mother} are not spouse. {child.mother} is mother of {child}"
                )
        else:
            if (
                child.father is not None
                and self.relation_with(child.father) != Relation.HUSBAND
            ):
                raise InvalidRelationError(
                    f"{child.father} and {self} are not spouse. {child.father} is father of {child}"
                )

        self.__add_relation(child, child_rel)
        child.__add_relation(self, parent_rel)

    def add_parent(self, parent: Person) -> None:
        parent.add_child(self)

    def add_spouse(self, other: Person) -> None:
        self.__validate_is_Person_object(other)

        if self.__gender == other.__gender:
            raise ValueError("Gender is same")

        if self.__gender == "m":
            husband = self
            wife = other
        else:
            husband = other
            wife = self

        husband.__add_relation(wife, Relation.WIFE)
        wife.__add_relation(husband, Relation.HUSBAND)

    def __repr__(self) -> str:
        return f"P{self.id}({self.name})"


class Lineage:
    def __init__(self) -> None:
        self._graph = DiGraph()
        self.__counter = -1

    def __new_id(self) -> int:
        self.__counter += 1
        return self.__counter

    # TODO Add id in graph node and person in node's attribute for faster search by id
    def add_person(self, name: str, gender: str) -> Person:
        person = Person(self._graph, self.__new_id(), name, gender)
        self._graph.add_node(person)

        return person

    def remove_person(self, person: Person) -> None:
        person.self_remove()

    def find_person_by_id(self, id: int) -> Person | None:
        for person in self.all_persons():
            if person.id == id:
                return person

    def find_person_by_name(self, name: str) -> list[Person]:
        name = name.lower()
        found = []
        for person in self.all_persons():
            if name in person.name.lower():
                found.append(person)
        return found

    def all_persons(self) -> list[Person]:
        return list(self._graph.nodes)

    def all_relations(self) -> list[(Person, Person, Relation)]:
        relations = []
        for p1, p2, relation in self._graph.edges.data():
            relations.append((p1, p2, relation[Relation]))
        return relations

    def all_unique_relations(self) -> list[(Person, Person, Relation)]:
        relations = []
        rel_set = []
        for p1, p2, relation in self._graph.edges.data():
            if (p1, p2) not in rel_set and (p2, p1) not in rel_set:
                rel_set.append((p1, p2))
                relations.append((p1, p2, relation[Relation]))

        return relations

    def shortest_path(self, start, stop):
        return networkx.shortest_path(self._graph, start, stop)

    def save_to_file(self, filename: Path | str) -> None:
        data = {
            "headers": {
                "persons": ["id", "name", "gender"],
                "relations": ["id1", "id2", "relation"],
            },
            "persons": [],
            "relations": [],
        }

        for person in self.all_persons():
            data["persons"].append([person.id, person.name, person.gender])

        for p1, p2, relation in self.all_relations():
            data["relations"].append([p1.id, p2.id, relation.name])

        with open(filename, "w") as f:
            json.dump(data, f, indent=0, separators=(",", ":"))

    @classmethod
    def load_from_file(cls, filename: Path | str) -> Lineage:
        lineage = cls()

        with open(filename) as f:
            data = json.load(f)
            persons_data = data["persons"]
            relations_data = data["relations"]

            persons_dict: dict[int, Person] = {}
            for prev_id, name, gender in persons_data:
                try:
                    # Here persons new id can be different or
                    # Lineage counter should be made equal to the highest id assigned plus 1
                    person = lineage.add_person(name, gender)

                    persons_dict[prev_id] = person
                except Exception:
                    pass

            for prev_id1, prev_id2, relation in relations_data:
                try:
                    prev_id1, prev_id2 = int(prev_id1), int(prev_id2)
                    persons_dict[prev_id1]._Person__add_relation(
                        persons_dict[prev_id2], Relation[relation]
                    )

                except Exception:
                    pass

        return lineage



import json
import requests
from pathlib import Path
from threading import Thread


LINEAGE_HOME = Path().home() / ".lineage"
LINEAGE_CONFIG_DIR = LINEAGE_HOME / ".config"
LINEAGE_AUTOSAVE_DIR = LINEAGE_HOME / "autosave"
alternate_spells_web_url = (
    "https://aqdasak.github.io/lineage_list/alternate_spells.json"
)


alternate_spells = []

config = {
    "print_all_ancestors": False,
    "print_id_with_person": False,
    "print_id_with_parent": False,
    # Possible: 0(Expand female only), 1(Expand male only), 2(Complete)
    "print_expanded_tree": 2,
    "print_spouse_in_tree": True,
}


def load_config():
    """
    Load the config from storage
    """

    file = LINEAGE_CONFIG_DIR / "lineage_config.json"

    global config
    try:
        with open(file) as f:
            config.update(json.load(f))
            if not isinstance(config.get("print_all_ancestors"), bool):
                config["print_all_ancestors"] = False
            if not isinstance(config.get("print_id_with_person"), bool):
                config["print_id_with_person"] = False
            if not isinstance(config.get("print_id_with_parent"), bool):
                config["print_id_with_parent"] = False
            if not 0 <= config.get("print_expanded_tree") <= 2:
                config["print_expanded_tree"] = 2
            if not isinstance(config.get("print_spouse_in_tree"), bool):
                config["print_spouse_in_tree"] = True
    except Exception:
        pass


def save_config():
    """
    Save the config to storage
    """

    file = LINEAGE_CONFIG_DIR / "lineage_config.json"

    with open(file, "w") as f:
        json.dump(
            # Here dict is built so that any extra key-value loaded
            # does not get save again
            {
                "print_all_ancestors": config["print_all_ancestors"],
                "print_id_with_person": config["print_id_with_person"],
                "print_id_with_parent": config["print_id_with_parent"],
                "print_expanded_tree": config["print_expanded_tree"],
                "print_spouse_in_tree": config["print_spouse_in_tree"],
            },
            f,
        )


def load_alternate_spells():
    """
    Load alternate_spells from storage
    """

    file = LINEAGE_CONFIG_DIR / "alternate_spells.json"
    global alternate_spells
    try:
        with open(file) as f:
            # If `[:]` is not used, new variable is created and data is
            # not written in global variable. Other solution is using `+=`
            alternate_spells[:] = json.load(f)
    except Exception:
        pass


def fetch_alternate_spells_from_web():
    """
    Download alternate_spells from the web, update the currently using alternate_spells and save to storage
    """

    def _fetch_alternate_spells_from_web():
        global alternate_spells
        file = LINEAGE_CONFIG_DIR / "alternate_spells.json"

        try:
            result = requests.get(alternate_spells_web_url).text
            if result:
                alternate_spells[:] = json.loads(result)
                with open(file, "w") as f:
                    f.write(result)

        except Exception:
            pass

    Thread(target=_fetch_alternate_spells_from_web, daemon=True).start()


def initialize_save_directories():
    """Create the directories if not already present"""

    LINEAGE_HOME.mkdir(parents=True, exist_ok=True)
    LINEAGE_AUTOSAVE_DIR.mkdir(parents=True, exist_ok=True)
    LINEAGE_CONFIG_DIR.mkdir(parents=True, exist_ok=True)


def setup():
    """Setup the program state, load the config and required variables"""

    initialize_save_directories()
    load_config()
    load_alternate_spells()
    fetch_alternate_spells_from_web()



from __future__ import annotations
from collections import defaultdict
from datetime import datetime
import json
import os
from typing import Callable
from lineage_aq import Lineage, Person, Relation, InvalidRelationError
from sys import exit
from lineage_aq.config import (
    LINEAGE_HOME,
    config,
    save_config,
    setup,
)
from lineage_aq.search import advanced_search
from lineage_aq.my_io import (
    input_from,
    input_in_range,
    non_empty_input,
    print_blue,
    print_cyan,
    print_green,
    print_grey,
    print_heading,
    print_id_name_in_box,
    print_tree,
    print_yellow,
    print_red,
    take_input,
)

lineage_modified = False


def commands() -> dict[Callable, str]:
    """Interactive commands for corresponding functions"""

    return {
        add_new_person: "new",
        add_parent: "addp",
        add_children: "addc",
        add_spouse: "adds",
        edit_name: "edit",
        remove_person: "rmperson",
        remove_relation: "rmrel",
        find: "find",
        show_tree: "tree",
        toggle_print_all_ancestors: "ta",
        toggle_print_id_with_person: "tid",
        toggle_print_id_with_parent: "tpid",
        toggle_print_expanded_tree: "texp",
        toggle_print_spouse_in_tree: "ts",
        shortest_path: "sp",
        no_parent: "noparent",
        one_parent: "oneparent",
        all_persons: "showall",
        all_relations: "showallrel",
        save_to_file: "save",
        safe_exit: "exit",
        show_help: "help",
    }


def person_repr(person: Person, parent=False) -> str:
    """Returns the representation of person based on certain switches"""

    if config["print_id_with_person"] or (config["print_id_with_parent"] and parent):
        return f"P{person.id}({person.name})"
    else:
        return person.name


def add_new_person(lineage: Lineage):
    def get_person_by_id(id: str) -> Person | None:
        try:
            if id:
                return lineage.find_person_by_id(int(id))
        except Exception as e:
            print_red("Error in ID", id)
            print_red(e)

    def find_same_name_persons(name: str) -> list[Person]:
        new_person_name = name.lower()
        all_persons = {
            person: person.name.lower().split(" ") for person in lineage.all_persons()
        }

        found = []
        for name in new_person_name.split(" "):
            found2 = []
            for person, fullname in all_persons.items():
                if name in fullname:
                    found2.append(person)

            # name may be a common name
            if len(found2) > 8:
                found2 = []
            found += found2

        # Fullname exact match
        for person in lineage.all_persons():
            if new_person_name == person.name.lower():
                found.append(person)

        return list(set(found))

    def whether_to_continue_if_found(name: str) -> bool:
        same_name_persons = find_same_name_persons(name)
        if same_name_persons:
            print_blue("\nPeople found with same name.")
            for person in same_name_persons:
                print_grey("-" * 50)
                _print_person_details(person)

            inp = input_from(
                "Do you want to continue to add new person? [y/n]: ",
                ("y", "n", "yes", "no"),
            )
            print()
            return True if inp in ("y", "yes") else False
        return True

    print_heading("ADD NEW PERSON")
    name = non_empty_input("Input name: ")
    if not whether_to_continue_if_found(name):
        return

    gender = input_from("Input gender (m/f): ", ("m", "f"))
    parent1 = get_person_by_id(take_input("Input ID of parent1 or leave blank: "))
    parent2 = get_person_by_id(take_input("Input ID of parent2 or leave blank: "))

    person = lineage.add_person(name, gender)
    try:
        if parent1:
            person.add_parent(parent1)
        if parent2:
            person.add_parent(parent2)
        print_yellow("Person added successfully")
    except ValueError as e:
        print_red("Parent2 not added")
        print_red(e)
    _print_person_details(person)

    global lineage_modified
    lineage_modified = True


def edit_name(lineage: Lineage):
    print_heading("EDIT NAME")
    person = lineage.find_person_by_id(int(non_empty_input("Enter ID of person: ")))
    print_cyan("Current name:", person.name)
    person.name = non_empty_input("Enter new name: ")
    _print_person_details(person)

    global lineage_modified
    lineage_modified = True


def add_parent(lineage: Lineage):
    print_heading("ADD PARENT")
    person = lineage.find_person_by_id(int(non_empty_input("Enter ID of person: ")))
    parent = lineage.find_person_by_id(int(non_empty_input("Enter ID of parent: ")))
    person.add_parent(parent)
    _print_person_details(person)

    global lineage_modified
    lineage_modified = True


def add_children(lineage: Lineage):
    print_heading("ADD CHILDREN")
    person = lineage.find_person_by_id(int(non_empty_input("Enter ID of person: ")))
    children = non_empty_input("Enter comma separated IDs of children: ")
    # Using dict as ordered set
    children = {i.strip(): "" for i in children.split(",")}
    children.pop("", None)

    for child_id in children:
        try:
            child = lineage.find_person_by_id(int(child_id))
            if child is None:
                print_red(f"ID={child_id}: ID is not present")
            else:
                person.add_child(child)
        except InvalidRelationError as e:
            print_red(f"ID={child_id}:", e)
        except Exception as e:
            print_red(e)
    _print_person_details(person)

    global lineage_modified
    lineage_modified = True


def add_spouse(lineage: Lineage):
    print_heading("ADD SPOUSE")
    person1 = lineage.find_person_by_id(int(non_empty_input("Enter ID I of person: ")))
    person2 = lineage.find_person_by_id(int(non_empty_input("Enter ID II of person: ")))
    try:
        person1.add_spouse(person2)
        print_yellow("Added successfully")
        _print_person_details(person1)
    except Exception as e:
        print_red("Something went wrong")
        print_red(e)

    global lineage_modified
    lineage_modified = True


def remove_person(lineage: Lineage):
    def is_any_relative_present(person) -> bool:
        for _, relatives in person.relatives_dict().items():
            if len(relatives) > 0:
                return True
        return False

    print_heading("REMOVE PERSON")
    person = lineage.find_person_by_id(int(non_empty_input("Enter ID of the person: ")))

    if is_any_relative_present(person):
        print_red("Relative(s) are present. First remove relations.")
    else:
        lineage.remove_person(person)
        print_cyan("Person removed")

    global lineage_modified
    lineage_modified = True


def remove_relation(lineage: Lineage):
    print_heading("REMOVE RELATION")
    person = lineage.find_person_by_id(int(non_empty_input("Enter ID I of person: ")))
    relative = lineage.find_person_by_id(
        int(non_empty_input("Enter ID II of person: "))
    )
    person.remove_relative(relative)
    print_cyan("Relation removed")
    _print_person_details(person)

    global lineage_modified
    lineage_modified = True


def _print_person_details(person: Person):
    def print_person(person: Person | list[Person], end="\n"):
        if isinstance(person, list):
            for p in person[:-1]:
                print_person(p, end=", ")
            print_person(person[-1])
        else:
            print_cyan(person_repr(person), end=end)

    father = person.father
    mother = person.mother
    sons = person.sons
    daughters = person.daughters
    husband = person.husband
    wife = person.wife

    if father:
        print_blue("Father:\t ", end="")
        print_cyan(person_repr(father, parent=True), end="")

        if config["print_all_ancestors"]:
            p = father.father
            while p:
                print_cyan(" -> ", end="")
                print_person(p, end="")
                p = p.father
        print()

    if config["print_all_ancestors"] and father and mother:
        # Space to reduce clutter
        print()

    if mother:
        print_blue("Mother:\t ", end="")
        print_cyan(person_repr(mother, parent=True), end="")

        if config["print_all_ancestors"]:
            p = mother.father
            while p:
                print_cyan(" -> ", end="")
                print_person(p, end="")
                p = p.father
        print()

    print_id_name_in_box(str(person.id), person.name)

    if husband:
        print_blue("Husband: ", end="")
        print_person(husband)

    if wife:
        print_blue("Wife:\t ", end="")
        print_person(wife)
    if sons:
        print_blue("Son:\t ", end="")
        print_person(sons)
    if daughters:
        print_blue("Daughter:", end="")
        print_person(daughters)

    brother = []
    sister = []
    if father:
        brother += father.sons
        sister += father.daughters
    if mother:
        brother += mother.sons
        sister += mother.daughters

    brother = sorted(set(brother) - {person}, key=lambda person: person.id)
    sister = sorted(set(sister) - {person}, key=lambda person: person.id)
    if brother or sister:
        print()
    if brother:
        print_blue("Brother: ", end="")
        print_person(brother)
    if sister:
        print_blue("Sister:  ", end="")
        print_person(sister)


def toggle_print_all_ancestors(_):
    global config
    config["print_all_ancestors"] = not config["print_all_ancestors"]

    command = commands()[toggle_print_all_ancestors]
    if config["print_all_ancestors"]:
        print_blue(f"{command}=ON")
        print("Ancestors will be shown")
    else:
        print_blue(f"{command}=OFF")
        print("Ancestors will not be shown")


def toggle_print_id_with_person(_):
    global config
    config["print_id_with_person"] = not config["print_id_with_person"]

    command = commands()[toggle_print_id_with_person]
    if config["print_id_with_person"]:
        print_blue(f"{command}=ON")
        print("ID will be shown for all")
    else:
        print_blue(f"{command}=OFF")
        print("ID will not be shown")


def toggle_print_id_with_parent(_):
    global config
    config["print_id_with_parent"] = not config["print_id_with_parent"]

    command = commands()[toggle_print_id_with_parent]
    if config["print_id_with_parent"]:
        print_blue(f"{command}=ON")
        print("Parents ID will be shown")
    else:
        print_blue(f"{command}=OFF")
        tid = commands()[toggle_print_id_with_person]
        print(f"Parents ID will not be shown, if {tid} is off")


def toggle_print_spouse_in_tree(_):
    global config
    config["print_spouse_in_tree"] = not config["print_spouse_in_tree"]

    command = commands()[toggle_print_spouse_in_tree]
    if config["print_spouse_in_tree"]:
        print_blue(f"{command}=ON")
        print("Spouse will be shown in tree")
    else:
        print_blue(f"{command}=OFF")
        print("Spouse will not be shown in tree")


def toggle_print_expanded_tree(_):
    global config

    config["print_expanded_tree"] = (config["print_expanded_tree"] + 1) % 3

    command = commands()[toggle_print_expanded_tree]
    if config["print_expanded_tree"] == 0:
        print_blue(f"{command}=FEMALE")
        print("Only females in tree will be expanded now")
    elif config["print_expanded_tree"] == 1:
        print_blue(f"{command}=MALE")
        print("Only males in tree will be expanded now")
    else:
        print_blue(f"{command}=ALL")
        print("Complete tree will be expanded now")


def _find_by_id(lineage: Lineage, id: int):
    person = lineage.find_person_by_id(id)
    if person:
        _print_person_details(person)
    else:
        print_red("ID not found")


def _find_by_name(lineage: Lineage, name: str):
    all_persons = {p.name: p for p in lineage.all_persons()}

    persons = advanced_search(name, all_persons.keys())
    if len(persons) == 0:
        print_red("Name not found")
        return

    for person_name in persons:
        person = all_persons[person_name]
        print_grey("─" * 50)
        _print_person_details(person)


def find(lineage: Lineage):
    print_heading("FIND PERSON")
    id_or_name = non_empty_input("Enter name or ID to search: ")
    if id_or_name.isdigit():
        _find_by_id(lineage, int(id_or_name))
    else:
        _find_by_name(lineage, id_or_name)


def all_persons(lineage: Lineage):
    print_heading("ALL PERSONS IN LINEAGE")
    for person in lineage.all_persons():
        print_cyan(person)


def all_relations(lineage: Lineage):
    print_heading("ALL RELATIONS IN LINEAGE")
    for relation in lineage.all_relations():
        print_cyan(relation)


def save_to_file(lineage: Lineage):
    global lineage_modified
    if not lineage_modified:
        print_red("No change since last save")
        inp = take_input("Do you want to save again [y/N]: ")
        if inp not in ("y", "yes"):
            print_grey("Not saved, since not required")
            return

    filename = (
        LINEAGE_HOME / f'lineage {datetime.now().strftime("%Y-%m-%d %H.%M.%S")}.json'
    )
    try:
        lineage.save_to_file(filename)
        print_green("Saved successfully at", filename)

        lineage_modified = False

    except Exception:
        print_red("Some error occured while saving file")


def autosave(lineage: Lineage):
    global lineage_modified
    if not lineage_modified:
        return

    filename = (
        LINEAGE_HOME
        / f'autosave/autosave-lineage {datetime.now().strftime("%Y-%m-%d %H.%M.%S")}.json'
    )
    try:
        lineage.save_to_file(filename)

        # lineage_modified = False
    except Exception:
        pass


def load_from_file() -> Lineage | None:
    def print_num_persons_and_relations(file):
        with open(file) as f:
            data: dict = json.load(f)

        num_persons = len(data["persons"])
        num_relations = len(data["relations"])

        print_grey(f"   [{num_persons}]", end="")
        print_grey(f"\t[{num_relations}]", end="")

    def print_all_files(files: list):
        padding = len(str(len(files)))
        print_yellow(" " * (padding - 1), end="")
        print("# ", "Filenames", " " * 23, "Persons  Relations")

        i = len(files)
        for file in reversed(files[1:]):
            print(f"{i:{padding}d}:", file.name, end="")
            print_num_persons_and_relations(file)
            print()
            i -= 1

        print(f"{1:{padding}d}:", files[0].name, end="")
        print_num_persons_and_relations(files[0])
        print_green(" (latest)")

    path = LINEAGE_HOME
    if not (path.exists() and path.is_dir()):
        print_red("Directory not found", path)
        return

    files = list(path.glob("*.json"))
    files.sort(reverse=True)

    if len(files) == 0:
        print_red("No saved file found in", path)
        return

    print_all_files(files)
    inp = int(input_in_range("Select the file to load: ", 1, len(files) + 1))

    return Lineage.load_from_file(files[inp - 1])


def safe_exit(lineage: Lineage):
    save_config()
    global lineage_modified
    if not lineage_modified:
        print_yellow("Exiting...")
        exit(0)

    inp = take_input(
        "You have unsaved data. Do you really want to exit without saving [y/N]: "
    )
    if inp in ("y", "yes"):
        autosave(lineage)
        exit(0)

    print_red("Exit aborted")


def shortest_path(lineage: Lineage):
    print_heading("SHORTEST PATH")
    person1_id = int(non_empty_input("Enter ID of I person: "))
    person2_id = int(non_empty_input("Enter ID of II person: "))

    sp = lineage.shortest_path(
        lineage.find_person_by_id(person1_id),
        lineage.find_person_by_id(person2_id),
    )

    for i in range(len(sp) - 1):
        print_cyan(sp[i], end=" ")
        print_blue(f"--{sp[i].relation_with(sp[i+1]).name}->", end=" ")
    print_cyan(sp[-1])

    print_cyan("Distance:", len(sp) - 1)


def _helper_no_and_one_parent(lineage: Lineage) -> tuple[set, set]:
    """Return set of persons having father and set of persons having mother"""

    relations = lineage.all_relations()
    rel = defaultdict(set)
    for p, _, r in relations:
        rel[r].add(p)

    return rel[Relation.FATHER], rel[Relation.MOTHER]


def no_parent(lineage: Lineage):
    print_heading("PERSONS HAVING NO PARENT")
    f, m = _helper_no_and_one_parent(lineage)

    allp = set(lineage.all_persons())
    no_parent = allp - f.union(m)

    if len(no_parent) == 0:
        print_red("All persons are having at least one parent")
        return

    no_parent = sorted(no_parent, key=lambda p: p.id)
    for i in no_parent:
        print(i)
    print_cyan("Total persons:", len(no_parent))


def one_parent(lineage: Lineage):
    print_heading("PERSONS HAVING SINGLE PARENT")
    f, m = _helper_no_and_one_parent(lineage)

    single_parent = f.union(m) - f.intersection(m)

    if len(single_parent) == 0:
        print_red("No person is having only single parent")
        return

    single_parent = sorted(single_parent, key=lambda p: p.id)
    for i in single_parent:
        print(i)

    print_cyan("Total persons:", len(single_parent))


def show_tree(lineage: Lineage):
    def _build_complete_tree(person: Person) -> dict[Person, dict]:
        tree = {}
        children = sorted(person.children, key=lambda p: p.id)
        for child in children:
            occurance[child] += 1
            tree[child] = _build_complete_tree(child)
        return tree

    def _build_male_expanded_tree(person: Person) -> dict[Person, dict]:
        tree = {}
        daughters = sorted(person.daughters, key=lambda p: p.id)
        for daughter in daughters:
            tree[daughter] = {}

        sons = sorted(person.sons, key=lambda p: p.id)
        for son in sons:
            occurance[son] += 1
            tree[son] = _build_male_expanded_tree(son)
        return tree

    def _build_female_expanded_tree(person: Person) -> dict[Person, dict]:
        tree = {}
        sons = sorted(person.sons, key=lambda p: p.id)
        for son in sons:
            tree[son] = {}

        daughters = sorted(person.daughters, key=lambda p: p.id)
        for daughter in daughters:
            occurance[daughter] += 1
            tree[daughter] = _build_female_expanded_tree(daughter)
        return tree

    print_heading("PRINT TREE")
    p_id = int(non_empty_input("Enter ID of person: "))
    person = lineage.find_person_by_id(p_id)

    occurance = defaultdict(int)
    if person is not None:
        if config["print_expanded_tree"] == 0:
            tree = {person: _build_female_expanded_tree(person)}
            print_blue("\nFemale expanded tree")
        elif config["print_expanded_tree"] == 1:
            tree = {person: _build_male_expanded_tree(person)}
            print_blue("\nMale expanded tree")
        else:
            tree = {person: _build_complete_tree(person)}
        # Since two persons can have same name, person_repr can't be used while building the tree dict
        print_tree(
            tree, occurance, person_repr, print_spouse=config["print_spouse_in_tree"]
        )

    else:
        print_red(f"ID {p_id} is not present")


def show_help(_=None, show_changes=False):
    def print_help(help: str, new=[]):
        for i, line in enumerate(help.split("\n")):
            if show_changes and i in new:
                print_green("(new)", end="")
                print_yellow(" " * 3, line)
            else:
                print_yellow(" " * 8, line)

    commands_help = f"""\
new:\t\tAdd new person
addp:\t\tAdd parent of a person
addc:\t\tAdd children of a person
adds:\t\tAdd spouse of a person
edit:\t\tEdit name of a person
find:\t\tFind and show matching person
tree:\t\tPrint tree of a person
sp:\t\tShortest path between two persons
rmrel:\t\tRemove relation between two persons
rmperson:\tRemove person from lineage
noparent:\tPersons whose no parent is present in lineage
oneparent:\tPersons whose only one parent is present in lineage
showall:\tShow all persons in lineage
showallrel:\tShow all relations in lineage
save:\t\tSave lineage to file
exit:\t\tExit the lineage prompt
help:\t\tShow this help

Type ID or name directly in the command field to search
Press {'<Ctrl>Z then Enter' if os.name == 'nt' else '<Ctrl>D'} in empty input to cancel
"""

    toggles_help = """\
ta:\t\tWhether to show all ancestors in person view
tid:\t\tWhether to show ID of all persons in person view and tree
tpid:\t\tWhether to show ID of parents in person view (if 'tid' is off)
texp:\t\tControl the expansion of tree
ts:\t\tWhether to show spouse in tree
"""

    print_yellow("USAGE: Type following commands to do respective action")
    print_help(commands_help, [6])
    print_yellow("\nTOGGLES/SWITCHES: Controls the output of other commands")
    print_help(toggles_help, [])


def _main():
    print_heading("LINEAGE")
    lineage = None

    try:
        inp = input_from(
            "Do you want to load lineage from file (y/n)? ", ("y", "n", "yes", "no")
        ).lower()
        if inp in ("y", "yes"):
            lineage = load_from_file()
    except (KeyboardInterrupt, Exception) as e:
        print_red("\nSomething went wrong")
        print(e)
        exit()

    show_help(show_changes=True)

    if lineage is None:
        print_yellow("Creating new lineage\n")
        lineage = Lineage()

    commands_fn = {v: k for k, v in commands().items()}

    while True:
        try:
            command = non_empty_input("# ").strip()
            command_ = command.replace(" ", "").lower()
            if command_ in commands_fn:
                commands_fn[command_](lineage)
            else:
                print_heading("FIND PERSON")
                if command.isdigit():
                    _find_by_id(lineage, int(command))
                else:
                    _find_by_name(lineage, command)

        except KeyboardInterrupt:
            print()
            try:
                safe_exit(lineage)
            except (KeyboardInterrupt, EOFError):
                print()

        except Exception as e:
            print_red(e)

        print_grey("─" * 50)


def main():
    setup()
    try:
        _main()
    except KeyboardInterrupt:
        exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Callable
from colorama import Fore as c, Style

from lineage_aq import Person


def arg_parse(*args):
    return " ".join(tuple(map(lambda x: str(x), args)))


def take_input(arg):
    arg = c.LIGHTGREEN_EX + arg + c.LIGHTYELLOW_EX
    inp = input(arg)
    print(c.RESET, end="")
    return inp


def non_empty_input(arg):
    arg = c.LIGHTGREEN_EX + arg + c.LIGHTYELLOW_EX
    while True:
        inp = input(arg)
        if inp != "":
            print(c.RESET, end="")
            return inp


def input_from(msg, from_: tuple, *, IGNORE_CASE=True):
    from2 = from_
    if IGNORE_CASE:
        from2 = [i.lower() for i in from_]

        while True:
            inp = non_empty_input(msg).lower()
            if inp in from2:
                return inp
            print_red(f"Warning: Input from {from_}")

    else:
        while True:
            inp = non_empty_input(msg)
            if inp in from2:
                return inp
            print_red(f"Warning: Input from {from_}")


def input_in_range(msg: str, a: int, b: int = None) -> float:
    if b is not None:
        if b <= a:
            raise ValueError("Upper bound should be greater than lower bound")
        lb = a
        ub = b
    else:
        if a <= 0:
            raise ValueError("Upper bound should be greater than 0")
        lb = 0
        ub = a

    while True:
        inp = non_empty_input(msg)
        try:
            inp = float(inp)
            if lb <= inp < ub:
                return inp
            print_red(f"Warning: Input range is [{lb},{ub})")
        except Exception:
            print_red("Warning: Please input a number")


def print_red(*args, end="\n"):
    end = c.LIGHTRED_EX + end + c.RESET
    print(c.LIGHTRED_EX + arg_parse(*args) + c.RESET, end=end)


def print_green(*args, end="\n"):
    end = c.LIGHTGREEN_EX + end + c.RESET
    print(c.LIGHTGREEN_EX + arg_parse(*args) + c.RESET, end=end)


def print_blue(*args, end="\n"):
    end = c.LIGHTBLUE_EX + end + c.RESET
    print(c.LIGHTBLUE_EX + arg_parse(*args) + c.RESET, end=end)


def print_yellow(*args, end="\n"):
    end = c.LIGHTYELLOW_EX + end + c.RESET
    print(c.LIGHTYELLOW_EX + arg_parse(*args) + c.RESET, end=end)


def print_cyan(*args, end="\n"):
    end = c.LIGHTCYAN_EX + end + c.RESET
    print(c.LIGHTCYAN_EX + arg_parse(*args) + c.RESET, end=end)


def print_grey(*args, end="\n"):
    end = c.LIGHTBLACK_EX + end + c.RESET
    print(c.LIGHTBLACK_EX + arg_parse(*args) + c.RESET, end=end)


def print_heading(*args):
    print()
    print(Style.BRIGHT + c.LIGHTMAGENTA_EX, end="")
    st = arg_parse(*args)
    print(st)
    print("-" * len(st), end="")
    print(c.RESET + Style.NORMAL)


def print_id_name_in_box(id: str, name: str):
    n_len = len(name)
    i_len = len(id)
    print_green("\t ╭" + "─" * (max(n_len, i_len) + 2) + "╮")

    print_blue("ID:\t", end=" ")
    print_green("│", end=" ")
    print_green(id, end=" ")
    print_green(" " * max(n_len - i_len, 0) + "│")

    print_blue("Name:\t", end=" ")
    print_green("│", end=" ")
    print_green(name, end=" ")
    print_green(" " * max(i_len - n_len, 0) + "│")
    print_green("\t ╰" + "─" * (max(n_len, i_len) + 2) + "╯")


def print_tree(
    tree: dict[Person, dict],
    occurance: dict[Person, int] = {},
    person_repr: Callable[[Person], str] = repr,
    print_spouse=True,
):
    """
    Print the dict of dict into tree

    Parameters
    ----------
    tree: dict[Person, dict]
        recursive dict containing parent as key and dict containing children
    occurance: dict[Person, int]
        contains the number of occurances of any person in the tree
    person_repr: Callable[[Person], str]
        function which represent string representation of the person to be printed

    Example
    -------
    ```
    tree = {
        "foo": {
            "bar": {"a": {"1": {"2": {}, "3": {}}}, "b": {}},
            "baz": {},
            "qux": {"c\nd": {}},
        },
    }

    print_tree(tree)
    ```

    Output:
    foo
    ├── bar
    │   ├── a
    │   │   └── 1
    │   │       ├── 2
    │   │       └── 3
    │   └── b
    ├── baz
    └── qux
        └── c⏎d
    """
    VERTICAL = "│   "
    HORIZONTAL = "─── "
    BRANCH = "├── "
    LAST_BRANCH = "└── "
    EMPTY = " " * len(BRANCH)
    LEN = len(BRANCH)

    def replace(st: str) -> str:
        return (
            st.replace(BRANCH, VERTICAL)
            .replace(LAST_BRANCH, EMPTY)
            .replace(HORIZONTAL, EMPTY)
        )

    def _print_tree(tree: dict[Person, dict], connector=""):
        print()
        if isinstance(tree, dict):
            connector = replace(connector) + BRANCH
            for i, (person, subtree) in enumerate(tree.items(), 1):
                if i == len(tree):
                    connector = replace(connector[:-LEN]) + LAST_BRANCH

                spouse = ""
                if print_spouse:
                    if person.gender == "m":
                        spouse = person.wife
                    else:
                        spouse = person.husband
                    if spouse:
                        spouse = [person_repr(s) for s in spouse]
                        spouse = "- " + ",".join(spouse)
                    else:
                        spouse = ""

                print(connector[LEN:], end="")
                if person in duplicates:
                    if duplicates[person]:
                        print_grey(person_repr(person), spouse, end="")
                        _print_tree(subtree, connector + c.LIGHTBLACK_EX)
                    else:
                        print(person_repr(person), spouse, end="")
                        duplicates[person] = True
                        _print_tree(subtree, connector)
                else:
                    print(person_repr(person), spouse, end="")
                    _print_tree(subtree, connector)

    # When value is False for first time, print the person in regular color, and then print it in grey for all other occurances
    duplicates = {person: False for person, freq in occurance.items() if freq > 1}
    _print_tree(tree)



from lineage_aq.config import alternate_spells


def mark_alternate_spells_tokens(x: str) -> str:
    """
    Enclose the alternate_spells tokens contained in the given string by `()`.

    Example
    -------
    (global) alternate_spells = [
        ["ee", "i"],
        ["aa", "a"],
    ]

    >>> generate_formattable_str("Hadees")
    H(a)d(ee)s
    """

    found_tokens = []
    i = 0
    for group in alternate_spells:
        for item in group:
            if item in x:
                found_tokens.append(item)
                # This is done so that there should be no replacement
                # in the already grabbed text
                x = x.replace(item, f"({i})")
                i += 1

    for i, item in enumerate(found_tokens):
        x = x.replace(f"({i})", f"({item})")
    return x


def create_variants(x: str) -> set[str]:
    """
    Create all the variants(permutations) by replacing the marked tokens by their alternates.

    Parameter
    ---------
    x: str
        String with marked tokens [output of function mark_alternate_spells_tokens()]

    Returns
    -------
    variants: set[str]
        set containing all variants

    Example
    -------
    (global) alternate_spells = [
        ["ee", "i"],
        ["aa", "a"],
    ]

    >>> create_variants("H(a)d(ee)s")
    {'Hadees', 'Haadees', 'Haadis', 'Hadis'}
    """

    variants = {x.replace("(", "").replace(")", "")}

    start = x.find("(")
    end = x.find(")")
    while start >= 0:
        spell = x[start + 1 : end]

        group_ = []
        for group in alternate_spells:
            if spell in group:
                group_ = group.copy()
                group_.remove(spell)
                break

        for sim in group_:
            a = x[:start] + sim + x[end + 1 :]
            variants |= create_variants(a)

        x = x.replace("(", "", 1).replace(")", "", 1)

        start = x.find("(")
        end = x.find(")")

    return variants


def advanced_search(search_term: str, search_space: list[str]) -> list[str]:
    """
    Search the given search_term in the given search_space, by generating several variants of the search_term.

    Parameters
    ----------
    search_term: str
        The term to be searched
    search_space: list
        List containing the data from which the term is to be searched

    Returns
    -------
    :list[str]
        List containing matching terms

    Example
    -------
    (global) alternate_spells = [
        ["ee", "i"],
        ["aa", "a"],
    ]

    >>> advanced_search("Hadees", ["Hadis", "Hadees", "Other"])
    ['Hadees', 'Hadis']
    """
    search_term = search_term.lower()
    variants = create_variants(mark_alternate_spells_tokens(search_term))

    result = []
    for variant in variants:
        for term in search_space:
            if variant in term.lower():
                result.append(term)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lineage_aq']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.5,<0.5.0', 'networkx>=2.8.5,<3.0.0', 'requests>=2.31.0,<3.0.0']

entry_points = \
{'console_scripts': ['lineage = lineage_aq.__main__:main']}

setup_kwargs = {
    'name': 'lineage-aq',
    'version': '1.4.0',
    'description': 'Interactive command line interface to create and edit lineage of families.',
    'long_description': '# Lineage\n\nGenealogy management software  \nStores lineage of family. Interactive command line interface to create and edit lineage of families. \n\n\n### Lineage\nnoun:\nthe series of families that somebody comes from originally  \nवंश, वंशावली\n\n\n# Install\n###### Recommended (To install pipx click [here](https://github.com/pypa/pipx#install-pipx))\n```\npipx install lineage-aq\n```\n\n###### or\n```\npip install lineage-aq\n```\n\n#### Or upgrade by:\n```\npipx upgrade lineage-aq\n```\n###### or\n```\npip install --upgrade lineage-aq\n```\n# Usage\n\n### Execute:\n```\nlineage\n```\n\n\n# Install from source\nPoetry is required. For installation click [here](https://python-poetry.org/docs/#installation).\n\n   Download the source and install the dependencies by running:\n  \n   ``` \n   git clone https://github.com/aqdasak/Lineage.git\n   cd Lineage\n   poetry install\n   ```\n\n### Run\nIn the source folder containing pyproject.toml, run\n```\npoetry shell\nlineage\n```\n\n> For Python3.7 compatible version click [here](https://github.com/aqdasak/Lineage-py37)\n',
    'author': 'Aqdas Ahmad Khan',
    'author_email': 'aqdasak@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/aqdasak/Lineage/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
